package edu.wpi.cs.dsrg.xmldb.xat.common.bindingtree;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dompattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.utils.BindingTable;
import java.util.*;
import java.lang.reflect.Array;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.*;

/**
 * This Class defines a tree representing a value in the Binding Table
 * It is composed of following type of nodes:
 * <pre>
 * Tag Node.
 * Attr Node.
 * ColumnName Node.
 * Text Node.
 * NavStep Node.
 * TextStep Node.
 * Function Node.
 * UnnestStep Node.
 * </pre>
 * It used for Nav/Tagger cancel out. It reused major code from DOMPattern.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */

public class BindingTree extends DOMPatternImp{
/**
 * The default BindingTree constructor.
 * @since 1.0
 */
public BindingTree() {
    super();
}
/**
 * This method takes a NavigationStep array and construct a BindingTree using it.
 * @param value The NavigationStep array to construct a BindingTree.
 * @since 1.0
 */
public BindingTree(NavigationStep[] value) {
	this();
	this.makeTree(value);
}
/**
 * The method make an indent format for output.
 * @return String The constructed indent formae string.
 * @param level The level of indent string.
 * @since 1.0
 */
String indent(int level) {
	StringBuffer space = new StringBuffer();
	for(int i = 0; i < level; i++)
		space.append(' ');
	return space.toString();
}
/**
 * This method traverse down through the given DOMPatternNode to find the first TagNode.
 * @return The first found DOMPatternNode.
 * @param node The DOMPatternNode searching for.
 * @since 1.0
 */
DOMPatternNode locateFirstTagNode(DOMPatternNode node) {
	if(node == null)
		return null;
	if(node instanceof TagNode)
		return node;
	DOMPatternNode[] children = node.getChildren();
	for(int i = 0; i < children.length; i++){
		DOMPatternNode result = locateFirstTagNode(children[i]);
		if(result != null)
			return result;
	}
	return null;
}
/**
 * Locate an open tag for a specified tag name within given NavigationStep.
 * The open tag is the "tag" has no "<>".
 * @return int -1 means not found.
 * @param value The NavigationStep array to locate the tag.
 * @param left The left searching bound.
 * @param right The right searching bound.
 * @param tag The tag name searching for.
 * @since 1.0
 */
int locateOpenTag(NavigationStep[] value, int left, int right, String tag) {
	if(value == null) 
		return -1;
	if(right >= value.length)
		right = value.length-1;
	if(left >= value.length || right < 0)
		return -1;
	if(left < 0)
		left = 0;
	if(tag == null)
		return -1;
		
	for(int istep = left; istep <=right; istep++){
		if(value[istep] instanceof TagStep){
			if(tag.equals(BindingTable.getTagValue(value[istep].getStepValue())))
				return istep;			
		}
	}
	return -1;
}
/**
 * Construct a DOMPatternNode according to the Step's type.
 * Tag Node.			TagStep.
 * Attr Node.			AttStep.
 * ColumnName Node.		VarStep.
 * Text Node.			StringStep.
 * Step Node.			NodeStep.
 * 	- Text Step Node.			TextStep.
 * 	- Function Step Node.		?(FunctionStep). #so far there is no such step.
 * 	- Unnest Step Node.		NodeStep(UnnestStep).
 * @return The result DOMPatternNode.
 * @param step The type of NavigationStep.
 * @since 1.0
 */
DOMPatternNode makeNode(NavigationStep step) {

    DOMPatternNode node = null;

    if (step == null)
        return null;

    if (step instanceof TagStep) {
        node = new TagNodeImp(BindingTable.getTagValue(step.getStepValue()));
    } else
        if (step instanceof AttStep) {
            node = new AttributeNodeImp(step.getStepValue());
        } else
            if (step instanceof VarStep) {
                node = new ColumnNameNodeImp(step.getStepValue());
            } else
                if (step instanceof StringStep) {
                    node = new TextNodeImp(step.getStepValue());
                } else
                    if (step instanceof NodeStep) {
                        node = new StepNodeImp(step.getStepValue());
                    } else
                        if (step instanceof TextStep) {
                            node = new TextNodeImp(step.getStepValue());
                        } else
                            if (step instanceof FunctionStep) {
                                node = new FunctionNodeImp(step.getStepValue());
                            } else
                                if (step instanceof UnnestStep) {
                                    node = new UnnestNodeImp(step.getStepValue());
                                } else
                                    if (step instanceof DescendantStep) {
                                        node = new StepNodeImp(step.getStepValue());
                                    } else {
                                        // unknow type of step.
                                        System.err.println(
                                            "ERROR! unknown type of NavigationStep in BindingTable.makeNode()");
                                    }
    return node;
}
/**
 * Construct a tree using given value array.
 * Following types of nodes are considered in the tree, and their mapping to navigationstep.
 * Tag Node.			TagStep.
 * Attr Node.			AttStep.
 * ColumnName Node.		VarStep.
 * Text Node.			StringStep.
 * Step Node.			NodeStep.
 * 	Text Step Node.			TextStep.
 * 	Function Step Node.		?
 * 	Unnest Step Node.		NodeStep.
 *
 * The value will be processed from the back to the front.
 * When we see the TagStep, we have to find both the open tag and the close tag.
 * @return void
 * @param value The NavigationStep array used to construct the Tree.
 * @since 1.0
 */
void makeTree(NavigationStep[] value) {
    if (value == null || value.length == 0)
        return;
    // process from the back to the front.
    // This is a recursive processing for the values start from lborder to rborder.
    this.rootNode = (RootNode) makeTree(value, 0, value.length - 1);
}
/**
 * This method recursively process the list of NavigationStep and translate it back to BindingTree.
 * When we see the TagStep, we have to find both the open tag and the close tag.
 * Tag Node.			TagStep.
 * Attr Node.			AttStep.
 * ColumnName Node.		VarStep.
 * Text Node.			StringStep.
 * Step Node.			NodeStep.
 * 	Text Step Node.			TextStep.
 * 	Function Step Node.		?
 * 	Unnest Step Node.		NodeStep.
 * @return DOMPatternNode The constructed tree.
 * @param value The NavigationStep value used to construct the BindingTree.
 * @param left The left bound of searching within NavigationStep array.
 * @param left The right bound of searching within NavigationStep array. 
 * @since 1.0
 */
DOMPatternNode makeTree(NavigationStep[] value, int left, int right) {
	DOMPatternNode root = new RootNodeImp();
	if(value == null || value.length == 0)
		return root;
	if(right < 0 || left < 0)
		return root;
	if(right < left)
		return root;
	if(right == left){
		// make one node and return it.
		DOMPatternNode node = makeNode(value[left]);
		this.addNode(node, root);
		return root;
	}
	// process from the back to the front.
	// If the node is not a closeTag, then make node, and link it to the root.
	// if the node is a closeTag, got its open Tag, make a node, and recursively
	// process it.
	DOMPatternNode cur = root;
	for(int istep = right; istep >=left; istep --){
		if(value[istep] instanceof TagStep){
			// locate the open tag.
			int iopen = locateOpenTag(value, left, right, BindingTable.getTagValue(value[istep].getStepValue()));
			// make the node for this tag.
			DOMPatternNode node = makeNode(value[istep]);
			// add it to the cur.
			this.addNode(node, cur);
			// recursively dive in to make the content of that tag as a subtree.
			DOMPatternNode subroot = makeTree(value, iopen+1, istep-1);
			// connect two trees together.
			if(subroot.getChildren().length != 0)
				this.addNode(subroot.getChild(0), node);
			// modify the istep to jump to the beginning of the open tag.
			istep = iopen;
		}else{
			// make corresponding Node.
			DOMPatternNode node = makeNode(value[istep]);
			// add it to the cur.
			this.addNode(node, cur);
			cur = node;
		}
	}
	return root;
}
/**
 * Minimize the BindingTree to do the cancel cout.
 * @return void
 * @since 1.0
 */
public void reduce() {
  this.reduce(this.getRoot());
}
/**
 * Minimize the BindingTree to do the cancel cout.
 * @return void
 * @param node The current node being processed.
 * @since 1.0
 */
private void reduce(DOMPatternNode node) {

    /** Algorithm
    * Traverse the tree Top-down.
    * . Find the first Tag.
    * . If there is no tag, then done.
    * . If the Tag node is the root node, then reset the root, recursively go down.
    * . else get the StepNode above it.
    * . Remove this Tag node.
    * . Get the reduced steps of the content of children of each of this Tag.
    * . Match the root, it could either be a Tag or a Step.
    * . If that's a Tag:
    * .   If NavStep is not TextNode, then match the step and tag value.
    * .   If Navstep is a TextNode, then it always match.
    * . If that's a navigate:
    * .   #text -- #text --> Remove outer #text.
    * .   #text -- step  --> do nothing.
    * .   step  -- #text --> no match. delete the subtree.
    * .   step  -- step  --> whether their values are match. no match, delete subtree.
    * At the end, convert the BindingTree back to the NavigationSteps. (reuse code).
    */

    // locate the first TagNode.
    DOMPatternNode tagNode = locateFirstTagNode(node);

    while (tagNode != null) { // if no tagnode found, we are done.

        DOMPatternNode[] children = tagNode.getChildren();

        // get the parent of this Tagger
        DOMPatternNode parentNode = tagNode.getParent();

        // depends on the type of the parent Node check the reducability of the current node.
        if (parentNode instanceof StepNode) {
            if ("UNNEST".equals(parentNode.getTagName())) {
                // delete the unnest and find again.
                this.removeNode(parentNode);
            } else {
                // get children node, and check their type.
                children = tagNode.getChildren();
                for (int ichild = 0; ichild < children.length; ichild++) {
                    if (children[ichild] instanceof TagNode) {
                        // if that's a tagger.
                        // the tagger's value has to match the step.
                        // otherwise delete this children.
                        if (!children[ichild].getTagName().equals(parentNode.getTagName())) {
                            // delete this subtree.
                            this.removeSubtree(children[ichild]);
                        }
                    } else
                        if (children[ichild] instanceof StepNode) {
                            // then the step must match.
                            if (!children[ichild].getTagName().equals(parentNode.getTagName())) {
                                // delete this subtree.
                                this.removeSubtree(children[ichild]);
                            }
                        }
                }
                // delete the found TagNode.
                // delete the Step.
                this.removeNode(tagNode);
                this.removeNode(parentNode);
            }
            tagNode = locateFirstTagNode(node);
        } else
            if (parentNode instanceof TextNode) {
                // get children node, and check their type.
                children = tagNode.getChildren();
                for (int ichild = 0; ichild < children.length; ichild++) {
                    if (children[ichild] instanceof TagNode) {
                        // then that's always a match.
                        // do nothing.
                    } else
                        if (children[ichild] instanceof StepNode) {
                            // then that's always a match.
                            // do nothing.
                        }
                }
                // delete the found TagNode.
                // delete the Step.
                this.removeNode(tagNode);
                this.removeNode(parentNode);
                tagNode = locateFirstTagNode(node);
            } else {
                // recursively dive into the tree for the tagNode's children.
                for (int ichild = 0; ichild < children.length; ichild++) {
                    this.reduce(children[ichild]);
                }
                // then terminate the while loop.
                tagNode = null;
            }
    }

}
/**
 * Remove the subtree identified by the given subroot.
 * @return void
 * @param subroot The root of the subtree removing.
 * @since 1.0
 */
public void removeSubtree(DOMPatternNode subroot) {
    DOMPatternNode parent = subroot.getParent();
    if (parent != null) {
        parent.deleteChild(subroot);
        subroot.setParent(null);
    }
}
/**
 * Convert current BindingTree as an Array of NavigationStep.
 * @return NavigationStep[] The convert result.
 * @since 1.0
 */
public NavigationStep[] toNavigationStep() {

	Vector  steps= this.toVector(this.getRoot());
	
    /* now we will add in the new binding to the vartable*/
	int size = steps.size();
    NavigationStep[] newBindingNav = new NavigationStepImp[size];
	for ( int i=0; i<size; i++)
	{
		Object temp = steps.elementAt(i);
		if ( temp instanceof String )
		{
			String tag = (String)temp;
			NavigationStep tempStep = null;
			if(tag.charAt(0) == '<'){// check if that's a tagger.
				tempStep = new TagStepImp(tag);
			}else{
				tempStep = new StringStepImp((String)temp);
			}
			newBindingNav[i] = tempStep;
			}
		else if ( temp instanceof ColumnName )
		{
			VarStep vartemp = new VarStepImp(  ((ColumnName)temp).getColumnName());
			newBindingNav[i] = vartemp;
		}
		else if (temp instanceof StepNode){
			newBindingNav[i] = new NodeStepImp(((StepNode)temp).getTagName());
		}else if (temp instanceof TextNode){
			newBindingNav[i] = new TextStepImp();
		}
	}
	return newBindingNav;
}
/**
 * Convert this BindingTree into a String.
 * @return String The result of converting.
 * @since 1.0
 */
public String toString() {
	return toString(this.rootNode, 0);
}
/**
 * Display this the tree with given DOMPatternNode as root.
 * @return String The convert result
 * @param node The root node of the tree.
 * @param level The depth level we will look at.
 * @since 1.0
 */
public String toString(DOMPatternNode node, int level) {
    // output the root.
    StringBuffer result = new StringBuffer();
    result.append(indent(level) + node.toString());
    // output the children.
    DOMPatternNode[] children = node.getChildren();
    if (children != null) {
        if (children.length != 0) {
            result.append("(\n");
            for (int i = 0; i < children.length; i++) {
                result.append(indent(level) + toString(children[i], level+1) + "\n");
            }
            result.append(indent(level) + ")");
        }
    }
    return result.toString();
}
/**
 * Converts the tree into a ListPattern by extracting the TagName and TagValue.
 * @return Vector The convert result.
 * @param node The root node of the tree converting for.
 * @since 1.0
 */
Vector toVector(DOMPatternNode node) {

    Vector result = new Vector();
    if (node == null)
        return result;

    DOMPatternNode[] children = node.getChildren();

    if (node instanceof RootNodeImp) {
        //handle the RootNode
        for (int i = 0; i < children.length; i++)
            result.addAll(toVector(children[i]));
    } else
        if (node instanceof TagNodeImp) {
            //handle Tag node

            // if the parent of this node and previous node are the same,
            // close the tag
            StringBuffer openTag = new StringBuffer();

            openTag.append("<" + node.getTagName());

            for (int i = 0; i < children.length; i++) {
                if (children[i] instanceof AttributeNodeImp) {

                    openTag.append(" " + children[i].getTagName() + "=\"");

                    DOMPatternNode attValue = children[i].getChild(0);

                    if (attValue instanceof TextNode) {
                        // if that's a text node, put the value here.
                        openTag.append(((TextNode) attValue).getTagValue());
                    } else
                        if (attValue instanceof ColumnNameNode) {
                            // if that's a column name node, put special varstep here.
                            //handle Column Name node
                            //get the first element of the Navigation Step Array
                            // represendint the column Name
                            Object o = Array.get(attValue.getTagValue(), 0);

                            String stepValue = ((NavigationStep) o).getStepValue();

                            result.add(openTag.toString());
                            result.add(new ColumnNameImp(stepValue));
                            openTag = new StringBuffer();
                        }
                    openTag.append("\"");
                }
            }

            // all the attributes have been printed, how print ">"
            openTag.append(">");

            result.add(openTag.toString());

            //Continue with the rest of the tree
            for (int i = 0; i < children.length; i++) {
                if (!(children[i] instanceof AttributeNodeImp))
                    result.addAll(toVector(children[i]));
            }

            //print the closing tag
            result.add("</" + node.getTagName() + ">");
        } else
            if (node instanceof ColumnNameNodeImp) {
                //handle Column Name node
                //get the first element of the Navigation Step Array
                // represendint the column Name
                Object o = Array.get(node.getTagValue(), 0);

                String stepValue = ((NavigationStep) o).getStepValue();

                result.add(new ColumnNameImp(stepValue));
                //handle subtree
                for (int i = 0; i < children.length; i++)
                    result.addAll(toVector(children[i]));
            } else
                if (node instanceof TextNodeImp) { //handle Text Node
                    //handle subtree
                    for (int i = 0; i < children.length; i++)
                        result.addAll(toVector(children[i]));

                    result.add(node);
                } else
                    if (node instanceof StepNodeImp) {
                        //handle subtree
                        for (int i = 0; i < children.length; i++)
                            result.addAll(toVector(children[i]));

                        result.add(node);
                    }

    return result;

}
}
