package edu.wpi.cs.dsrg.xmldb.xat.component.operatorfactory;

import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.FunArith;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.Pattern;
import edu.wpi.cs.dsrg.xmldb.xat.common.expression.Expression;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.streamoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
/**
 * <pre>
 * Abstract base class for all operator builders.  A handle to a
 * subclass of OperatorBuilder is obtained from the appropriate OperatorFactory using
 * <code>OperatorFactory.newOperatorBuilder();</code>
 * The OperatorBuilder is responsible for instantiating operators and returning handles to
 * the appropriate interface
 *
 * When adding methods to the builder to create new operators, they should be made abstract iff
 * the stream and nonstream logical semantics are different.  In most cases they will be.  Note:
 * newOperator methods cannot just be added to one builder or another, but rather all builders must be able
 * to provide that method.  This is because the outside only has a handle to this abstract base class, not 
 * any individual subclass.   
 * </pre>
 * @author: Brad Pielech
 */
public abstract class OperatorBuilder {
	protected static OperatorBuilder builder = null;
/**
 * OperatorBuilder constructor comment.
 */
protected OperatorBuilder() {
	super();
}
/**
 * Returns a handle to the operatorBuilder
 * This is a friend method so it can only be invoked by the factory
 * This is overwritten by the subclasses so returning null will not cause
 * problems.  The method would be called abstract, but you cannot have a static abstract
 * method in java.
 * @return edu.wpi.cs.dsrg.xmldb.xat.component.operatorfactory.OperatorBuilder
 */
static OperatorBuilder getOperatorBuilder(){
	return null;
}
/**
 * Create a new Aggregate
 * @return Aggregate
 * @param node XATNode
 */
public Aggregate newAggregate(XATNode node){
	return new AggregateImp(node);
}
/**
 * Create a new CartesianProduct
 * @return CartesianProduct
 * @param node XATNode
 */
public CartesianProduct newCartesianProduct(XATNode node){
	return new CartesianProductImp(node);
}
/**
 * Create a new Distinct
 * @return Distinct
 * @param node XATNode
 */
public Distinct newDistinct(XATNode node){
	return new DistinctImp(node);
}
/**
 * Create a new Expose
 * This doesnt really have a stream equivalent so it will be implemented here
 * @return Expose
 * @param node XATNode 
 */
public Expose newExpose(XATNode node){
	return new ExposeImp(node);
}
/**
 * Create a new FOR
 * This doesnt really have a stream equivalent so it will be implemented here
 * @return FOR
 * @param node XATNode
 */
public FOR newFor(XATNode node){
	return new FORImp(node);
}
/**
 * Create a new FOR
 * This doesnt really have a stream equivalent so it will be implemented here
 * @return FOR
 * @param node XATNode
 * @param tree XATTree
 */
public FOR newFor(XATNode node, XATTree tree){
	return new FORImp(node, tree);
}
/**
 * Create a new FunctionArith
 * This doesnt really have a stream equivalent so it will be implemented here
 * @return FunArith
 * @param node XATNode 
 */
public FunArith newFunArith(XATNode node){
	return new FunArith(node);
}
/**
 * Create a new Function
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.Function
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
public edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.Function newFunction(XATNode node){
	return new edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.FunctionImp(node);
}
/**
 * Insert the method's description here.
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupBy
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
public GroupBy newGroupBy(XATNode node){
	return new GroupByImp(node);
}
/**
 * Create a new LeftOuterJoin
 * @return LeftOuterJoin
 * @param node XATNode
 * @param exp Expression
 */
public LeftOuterJoin newLeftOuterJoin(XATNode node, Expression exp){
	return new LeftOuterJoinImp(node, exp);
}
/**
 * Create a new Merge
 * This doesnt really have a stream equivalent so it will be implemented here
 * @return Merge
 * @param node XATNode 
 */
public Merge newMerge(XATNode node){
	return new MergeImp(node);
}
/**
 * Create a new NameColumn
 * @return NameColumn
 * @param node XATNode
 */
public NameColumn newNameColumn(XATNode node){
	return new NameColumnImp(node);
}
/**
 * Create a new NavCollection operator 
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavCollection
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
public NavCollection newNavCollection(XATNode node){
	return (NavCollection)new NavCollectionImp(node);
}
/**
 * Create a new NavCollection operator 
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.NavUnnest
 * @param XATNode node the XATNode to associate with this operator
 */
public NavUnnest newNavUnnest(XATNode node){
	return (NavUnnest) new NavUnnestImp(node);
}
/**
 * Create a new OrderBy
 * @return OrderBy
 * @param node XATNode
 */
public OrderBy newOrderBy(XATNode node){
	return new OrderByImp(node);
}
/**
 * Create a new Project
 * @return Project
 * @param node XATNode
 */
public  Project newProject(XATNode node){
	return new ProjectImp(node);
}
/**
 * Create a new Select
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.Select
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 */
public Select newSelect(XATNode node, Expression exp){
	return (Select) new SelectImp(node, exp);
}
/**
 * Create a new Source
 * @return Source
 * @param node XATNode
 */
public Source newSource(XATNode node){
	return new SourceImp(node);
}
/**
 * Create a new Tagger
 * 
 * @return edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.Tagger
 * @param node edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode
 * @param pattern Pattern 
 */
public Tagger newTagger(XATNode node, Pattern pattern){
	return (Tagger) new TaggerImp(node, pattern);
}
/**
 * Create a new XMLUnion
 * @return XMLUnion
 * @param node XATNode
 */
public XMLUnion newXMLUnion(XATNode node){
	return new XMLUnionImp(node);
}
}
