package edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.Hashtable;
import java.util.Set;
import java.util.HashSet;

import edu.wpi.cs.dsrg.xmldb.xat.component.schemacleanup.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.XATNode;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.XATTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.InheritableVisitor;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.XATOperator;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.IfExpr;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.FOR;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.GroupBy;

/**
 * This class will call the SchemaVisitor to compute the Minimum Schema,
 * call the RequiredVisitor to compute the Required Columns, and 
 * do the cutting if the required columns not in produced and modified.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 */
 
public class CutVisitor extends InheritableVisitor
{
    /**
     * The hash table to keep the map between XATNode and its corresponding SchemaCleanUpNode.
     * @since 1.0     
     */
	Hashtable hash;
    /**
     * The visitor to computer the required schema of nodes in 
     * the given XATTree.
     * @since 1.0     
     */
    RequiredVisitor rvisitor;
    /**
     * The visitor to computer the minimum schema of nodes in 
     * the given XATTree.
     * @since 1.0     
     */   
    SchemaVisitor svisitor;
    /**
     * Indicates whether some of the node got cutted.
     * @since 1.0
     */
	public boolean cutted;
/**
 * CutVisitor Constructor.
 * @param  hash_init The hash table to keep the map between 
 * XATNode and its corresponding SchemaCleanUpNode.
 * @since 1.0
 * @see RequiredVisitor
 * @see SchemaVisitor
 */
public CutVisitor(Hashtable hash_init) {
    hash = hash_init;
    rvisitor = new RequiredVisitor(hash);
    svisitor = new SchemaVisitor(hash);
    cutted = false;
}
/**
 * Decide the given XATNode cuttable or not.
 * This method uses the SchemaCleanupNode of current XATNode to decide if current node
 * cuttable or not.
 * @return boolean True: Cuttable. False: Not cuttable.
 * @param scn The current SchemaCleanupNode.
 * @param scnparent The parent SchemaCleanupNode.
 * @since 1.0
 * @see SchemaCleanupNode
 */
private boolean cuttable(SchemaCleanupNode scn, SchemaCleanupNode scnParent) {
    // check the validity of the parameters.
    if (scn == null || scnParent == null) {
        return false;
    }

    Set s;
    if (scn.produced != null)
        s = new HashSet(scn.produced);
    else
        s = new HashSet();

    Set sp;
    if (scnParent.required != null)
        sp = new HashSet(scnParent.required);
    else
        sp = new HashSet();

    if (scn.allModified || scnParent.allModified) {
        return false;
    }

    if (scn.modified != null)
        s.addAll(scn.modified);

    if (scnParent.modified != null)
        sp.addAll(scnParent.modified);

    s.retainAll(sp);
    return s.isEmpty();
}
    public Object dispatch()
    {
	return null;
    }
    public Object dispatch(Object o)
    {
	return null;
    }
/**
 * Get the attribute "cutted".
 * @return boolean True: cuttable. False: not cuttable.
 * @since 1.0
 */
public boolean getCutted() {
	return this.cutted;
}
/**
 * Set the attribute "cutted" using given boolean value.
 * @return boolean True: cuttable. False: not cuttable.
 * @param cutvalue The boolean value indicate if the node is cuttable or not.
 * @since 1.0
 */

public boolean setCutted(boolean cutvalue) {
	return this.cutted=cutvalue;
}
/**
 * This method recursively traverse the XATTree to execute the function 
 * of the given InheritableVisitor.
 * @return void
 * @param node The XATNode going to start the traversing.
 * @param v The functional visitor.
 * @since 1.0
 */

private void traverseHelper(XATNode node, InheritableVisitor v) {
    v.visit(node);
    XATNode[] children = node.getChildren();
    for (int i = 0; i < children.length; i++) {
        traverseHelper(children[i], v);
    }
}
/**
 * This method traversed the XATTree to compute the Produce, consume, and minimun Schema.
 * @return void
 * @param tree The XATTree going to traverse.
 * @param requiredAtRoot The set of required columns at root.
 * @since 1.0
 * @see XATTree
 * @see PCMVisitor
 * @see MinSchemaVisitor
 * @see traverseHelper(XATNode, InheritableVisitor) traverseHelper
 */

private void traverseSubtree(XATTree tree, Set requiredAtRoot) {
    Hashtable localhash = new Hashtable();
    PCMVisitor pcmvisitor = new PCMVisitor(localhash);
    XATNode root = tree.getRoot();
    traverseHelper(root, pcmvisitor);
    SchemaCleanupNode scn = (SchemaCleanupNode) localhash.get(root);
    scn.schema = requiredAtRoot;
    root.setSchema(requiredAtRoot);
    Set s = new HashSet(scn.schema);
    s.removeAll(scn.produced);
    s.addAll(scn.consumed);
    scn.required = s;
    XATNode[] children = root.getChildren();
    MinSchemaVisitor rscvisitor = new MinSchemaVisitor(localhash);
    for (int i = 0; i < children.length; i++) {
        traverseHelper(children[i], rscvisitor);
    }
}
/**
 * Compute all consumed columns by a given Expression.
 * @return Object 
 * @param  node   The XATNode being cleaned.
 * @since 1.0
 * @see SchemaCleanupNode
 * @see traverseSubtree(XATTree, Set) traverseSubtree
 */
 
public Object visit(XATNode node) {

    SchemaCleanupNode scn = (SchemaCleanupNode) hash.get(node);
    svisitor.visit(node);
    rvisitor.visit(node);
    XATNode parent = node.getParent();
    XATNode[] children = node.getChildren();
    SchemaCleanupNode scnParent;
    if (parent == null) {
        scnParent = new SchemaCleanupNode();
        scnParent.modified = new HashSet();
        scnParent.required = scn.schema;
        scnParent.allModified = false;
    } else {
        scnParent = (SchemaCleanupNode) hash.get(parent);
        /* we don't want to change the degree of the parent node */
        if ((children.length == 1) && cuttable(scn, scnParent)) {
            /** 
             * we essentially want XATTree.deleteNode(node); but we don't
             * want to actually see the tree.  
             */
            XATNode[] parentschildren = parent.getChildren();
            for (int i = 0; i < parentschildren.length; i++) {
                if (parentschildren[i] == node) {
                    parent.deleteChild(node);
                    setCutted(true);
                    if (isDebug()) {
                        System.out.println("Cutted Node: " + node);
                    }
                    int l = children.length - 1;
                    for (int j = l; j >= 0; j--) {
                        parent.addChild(children[j], i);
                    }
                    parentschildren = parent.getChildren();
                    i += l;
                }
            }
            for (int i = 0; i < children.length; i++) {
                children[i].setParent(parent);
            }
            return null;
        }
    }

    // get the current operator.
    XATOperator op = node.getOperator();

    if (op instanceof FOR || op instanceof GroupBy || op instanceof IfExpr) {

        // get it's outer tree's schema.
        XATNode outerRoot = node.getChildren()[0];

        // disable the link of the parent node, hence it will not confuse the SchemaVisitor.
        outerRoot.setParent(null);

        // compute the outer tree's schema first.
        visit(outerRoot);

        // restore the parent link.
        outerRoot.setParent(node);

        // get the computed schema.
        SchemaCleanupNode outerScn = null;
        if (outerRoot != null)
            outerScn = (SchemaCleanupNode) hash.get(outerRoot);
        else {
            outerScn = new SchemaCleanupNode();
            outerScn.modified = new HashSet();
            outerScn.required = new HashSet();
            outerScn.allModified = false;
        }

        if (op instanceof FOR) {
            Set s = new HashSet(outerScn.schema);
            s.retainAll(outerScn.produced);
            traverseSubtree(((FOR) op).getXtreFunctions(), s);
        } else
            if (op instanceof GroupBy) {
                Set s = new HashSet(outerScn.schema);
                s.retainAll(outerScn.produced);
                traverseSubtree(((GroupBy) op).getXATTree(), s);
            } else
                if (op instanceof IfExpr) {
                    Set s = new HashSet(outerScn.schema);
                    s.retainAll(outerScn.produced);
                    traverseSubtree(((IfExpr) op).getThenXAT(), s);
                    traverseSubtree(((IfExpr) op).getElseXAT(), s);
                }
    }
    return null;
}

/**
 * Get the value of the propertity "DEBUG_SchemaCleanUp".
 * @return boolean True: With debug information. False: No debug information.
 * @since 1.0
 */
public static boolean isDebug() {
	return System.getProperty("DEBUG_SchemaCleanUp")!=null;
}
}
