package edu.wpi.cs.dsrg.xmldb.xat.component.xatrewrite;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.*;
import java.lang.reflect.*;

import edu.wpi.cs.dsrg.utils.*;
import edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.expression.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.dompattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.exception.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.xatnode.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.visitorpattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.updateoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.sqloperator.SelectImp;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.specialoperator.function.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.*;

/**
 * This class is the super class for all equivalence rule classes. 
 * <pre>
 * An object of this class is created with a Vector of Strings containing the order
 * of classes to look at when trying to find a rule for 2 operators. 
 *
 * Calls are then made to evaluateRules with 2 nodes that the user wishes to evalute. 
 * The TwoParamVisitor class from the visitorpattern package is used to determine the correct 
 * equivalence rule to use for a given pair of XATOperators. This is done by iterating over 
 * the classes provided in the classVector and calling the visit(Object, Object) method of each class. 
 * 
 * If a rule is not found in a class, the next class is looked at.  If no rules are found for a given
 * pair, an ReWriteRulesNoRulesFound exception is thrown.
 * 
 * Equivalence rules are defined in subclasses named after the parent operator.  ReWriteNavigationRules
 * defines the equivalence rules for Navigate Operators.  ReWriteOrderByRules defines the
 * equivalence rules for OrderBy Operators and so on. 
 * 
 * A rule is a method in the strucure of visit(XATOperator parent, XATOperator child).  The rule
 * for swapping a Navigate with a Source is defined in ReWriteNavigationRules.visit(Navigate, Source). 
 * Each rule will perform modififactions neccessary for the swap of the two operators if such a swap is
 * possible.  If the nodes can be swapped, the rule will do so. 
 * See each rule for the logic behind that rule.  Each rule returns a ReWriteResult object. 
 * This object will tell the calling program if the Operators were be swapped, is the parent in its
 * final position, and should either node be deleted.  
 * </pre>
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0 
 * @see  ReWriteResult
 */
 
public class ReWriteRules extends TwoParamVisitor {

    /** 
     * ReWriteResult that describes the results from performing the Equivalence Rule
     * @since 1.0
     */
    protected ReWriteResult result;
    /** 
     * Vector that contains the list and order of rule classes to look at.
     * @since 1.0
     */
    protected Vector classVector;
    /** 
     * BindingTable that is used to store the variables and the associated columnPaths
     * @since 1.0
     */
    protected BindingTable variableTable;
    /** 
     * Hashtable to keep the rule statistics of this rewrite rules
     * In the format of "method obj, Integer count".
     * This is an object for all the instances of this class.
     * @since 1.0
     */
    protected static Hashtable statis;
    /**
     * Default ReWriteRules constructor.
     * @since 1.0
     */
    public ReWriteRules() {

        result = new ReWriteResult();
        classVector = new Vector();
        variableTable = new BindingTable(); 

    }
/**
 * ReWriteRules constructor.
 * @param sourceList The list of operator types being considered to apply this rewrite rule.
 * @since 1.0  
 * @see BindingTable 
 */

public ReWriteRules(Vector sourceList) {
    result = new ReWriteResult();
    classVector = sourceList;
    variableTable = new BindingTable();

}
/**
 * Constructor that accepts a Vector sourceList that is set to classVector 
 * and a Hashtable that is set to variableTable. The result object is
 * instantiated.  
 * @param sourceList The list of operator types being considered to apply this rewrite rule.
 * @param BindingTable The binding table of given XAT.
 * @since 1.0  
 * @see BindingTable 
 */

public ReWriteRules(Vector sourceList, BindingTable varList) {

    result = new ReWriteResult();
    classVector = sourceList;
    variableTable = varList;

}
/**
 * This method puts the entryPoint and destination step arrays of a Navigate 
 * Operator into 1 bigger array to make it easier to manipulate.
 * @return NavigationStep[] The array for combined Navigation Steps.
 * @param entryPoint  The entry point of Navigate operator.
 * @param destination The destination of Navigate operator. 
 * @since 1.0
 * @see NavigationStep
 */
protected NavigationStep[] combineSteps(
    NavigationStep[] entryPoint,
    NavigationStep[] destination) {

    int combinedLength = 0;
    int entryLength = 0, destLength = 0;
    try {
        entryLength = entryPoint.length;
    } catch (NullPointerException e) {
        //entry point is null
    }
    try {
        destLength = destination.length;
    } catch (NullPointerException e) {
        //dest is null

    }
    combinedLength = entryLength + destLength;
    NavigationStep[] combinedPath = new NavigationStep[combinedLength];
    //put in all of the entryPoint steps	
    for (int x = 0; x < entryLength; x++) {
        if (entryPoint[x] != null) {
            combinedPath[x] = entryPoint[x];
        }
    }
    //now put in all the destination steps
    for (int y = entryLength; y < combinedLength; y++) {
        if (destination[y - entryLength] != null) {
            combinedPath[y] = destination[y - entryLength];
        }
    }

    return combinedPath;
}
/**
 * Replacing the child node with parent node.
 * <pre>
 * This method will alter the XATTree structue by copying the parent
 * node directly replace child. The child's children will be altered to
 * point to parent. This is primarily used to copy a node under the correct 
 * Join / Union branch.
 *
 * <b> Structure of tree </b>
 * parent = x2, child = x4
 *   
 *   	old structure
 *   	x1 -> x2 ---> x3 -> x4
 *   
 *   	new structure
 *   	x1 -> x2 ---> x3 -> Clone of x2 -> x4
 * <pre>
 * @return boolean The boolean value indicates operation finished successfully.
 * @param parent The parent node.
 * @param child  The child node.
 * @since 1.0
 * @see #swap(XATOperator, XATOperator) swap
 */
 
protected boolean copyNode(XATNode parent, XATNode child) {

    XATNode x1 = parent.getParent();
    // clone the parent.
    XATNode x2 = (XATNode) parent.clone();
    // remove all children of the clone.
    x2.setChildren(new XATNode[] {
    });

    XATNode x3 = child.getParent();
    XATNode x4 = child;

    //get the index of x4 of x3.
    int index = x3.getIndexOf(x4);

    //replace x4 with x2.
    x3.setChild(index, x2);
    x2.setParent(x3);

    //add x4 as children of x2.
    x2.addChild(x4);
    x4.setParent(x2);

    // done. tree is modified.
    return true;
}
/**
 * Convert an array of NavigationStep into String.
 * This method accepts an array of NavigationSteps and will convert these 
 * steps into one long string.
 * If a VarStep is found, it will be replaced by the actual column name(s) as found
 * in this.variableTable
 * The array is formatted with "/" between steps.  A "/" is added at the end of the string as well
 * @return String The result corresponding step string.
 * @param step NavigationStep from which to create a string.
 * @since 1.0
 * @see NavigationStep
 */
 
public String createString(NavigationStep[] step) {
    String returnString = new String();
    String stepValue = "";
    NavigationStep[] varValue;

    for (int i = 0; i < step.length; i++) {
        if (step[i] instanceof VarStep) {
            varValue = getVariableTable().getVariableValue(step[i].getStepValue());
            for (int j = 0; j < varValue.length; j++) {
                stepValue = stepValue + varValue[j].getStepValue() + "/";
            }

        } else {
            stepValue = step[i].getStepValue() + "/";
        }

        returnString = returnString + stepValue;
        stepValue = "";

    }

    return returnString;
}
/**
 * Decide if an given Aggregate operator can be deleted or not.
 * The given Aggregrate operator can be deleted if either its parent or child 
 * is not a Tagger.
 * @return boolean True: if the aggregate can be deleted. False: otherwise.
 * @param param The node of Aggregrate operator.
 * @since 1.0
 * @see Tagger
 * @see Aggregate
 */
 
public boolean deleteAggregate(XATNode param) {
    try {
        XATOperator parent = param.getParent().getOperator();
        XATOperator child = param.getChild(0).getOperator();
        if ((parent instanceof Tagger) || (child instanceof Tagger)) {
            return false;
        } else {
            return true;
        }

    } catch (Exception e) {
        return false;
    }

}
/**
 * Decide if a Tagger operator can be deleted or not.
 * This method will iterate over a Tagger operator and determine if the
 * operator can be deleted or not.  A Tagger may be deleted when all of 
 * its nodes have been cancelout
 * @return boolean True: if the tagger can be deleted. False: otherwise.
 * @param rootNode The root of the DOMPattern to start iterating over.
 * @since 1.0
 * @see Tagger
 * @see DOMPatternNode
 * @see RootNodeImp
 * @deprecated
 */
public boolean deletePartialTagger(DOMPatternNode rootNode) {

    DOMPatternNode[] children = rootNode.getChildren();
    for (int i = 0; i < children.length; i++) {

        //if this is a rootNode, then we dont check for cancelout
        if (rootNode instanceof RootNodeImp) {
            return deleteWholeTagger(children[i]);
        } else {
            //check for the cancelout property
            if (rootNode.isCanceledOut()) {
                /** This may result in an incorrect tree */
                DOMPatternNode parent = rootNode.getParent();
                parent.deleteChild(rootNode);

            } else {
                return false;

            }

            return deleteWholeTagger(children[i]);
        }

    }

    return true;
}
/**
 * Decide if the Tagger can be deleted or not.
 * This method will iterate over a Tagger operator and determine if the
 * operator can be deleted or not.  A Tagger may be deleted when all of its nodes
 * have been cancelout.
 * @return boolean True: if the tagger can be deleted. False: otherwise.
 * @param rootNode The root of the DOMPattern to start iterating over.
 * @since 1.0
 * @see Tagger
 * @see RootNodeImp
 * @see DOMPatternNode
 * @deprecated
 */
 
public boolean deleteWholeTagger(DOMPatternNode rootNode) {

    DOMPatternNode[] children = rootNode.getChildren();
    for (int i = 0; i < children.length; i++) {

        //if this is a rootNode, then we dont check for cancelout
        if (rootNode instanceof RootNodeImp) {
            return deleteWholeTagger(children[i]);
        } else {
            //check for the cancelout property
            if (rootNode.isCanceledOut()) {

            } else {
                return false;

            }

            return deleteWholeTagger(children[i]);
        }

    }

    return true;
}
public Object dispatch() {
    return null;
}
    public Object dispatch(Object object) {
        return null;
    }
public Object dispatch(Object object1, Object object2) {
    /*
            System.out.println("In dispatch Object, Object in super class");
            System.out.println("Object 1 is: " + object1.toString());
            System.out.println("Object 2 is: " + object2.toString());
    */
    return null;
}
/**
 * Evaluate the rules over 2 XATOperators.  
 * <pr>
 * This method uses java.lang.reflect methods to instantiate the classes
 * in this.classVector.  The method iterates over the classes, instantiates the class,
 * then calls the visit function with the operators of parent and child.
 *
 * If no rule is found, the next class is instantiated and the process repeats itself.  
 * the function returns true if the function cannot be pushedDown further
 * otherwise it returns false
 * </pr>
 * @return ReWriteResult An object containing a summary of what was performed by the rule.
 * @param parent The parent node that we will try to find a rule for.
 * @throws ReWriteRulesNoRulesFound Exception thrown when the system has looked at
 * all classes in the classVector and is unable to find a rule for parent and child.
 * @since 1.0
 * @see ReWriteResult
 */
public ReWriteResult evaluateRules(XATNode parent)
    throws ReWriteRulesNoRulesFound {

    String className;
    Class ruleClass;
    ReWriteRules p1;
    /*
            System.out.println(
                "Trying to find a rule for: "
                    + parent.getOperator().toString()
                    + " and "
                    + child.getOperator().toString());
    */
    className = parent.getOperator().getClass().toString();

    for (int x = 0; x < classVector.size(); x++) {

        try {
            className = (String) classVector.elementAt(x);
            ruleClass = Class.forName(className);
            //create a new instance of the class from the classVector
            p1 = (ReWriteRules) ruleClass.newInstance();
            //set the variableTable of the new rule Class
            p1.setVariableTable(this.variableTable);
            //set the debug value for each new rule Class
            p1.setDebug(this.isDebug());
            //call the visit(Object, Object) of that class
            result = (ReWriteResult) p1.visit(parent.getOperator());
            //return the result if it is not null
            if (result != null) {
                //make the last rule class looked at the first class to make it faster
                String lastClass = (String) classVector.elementAt(x);
                classVector.removeElement(lastClass);
                classVector.insertElementAt(lastClass, 0);
                return result;
            } else {
                //otherwise look at the next class
            }
        } catch (Exception e) {
            System.out.println(e.toString());
        }
    }

    throw new ReWriteRulesNoRulesFound(
        "No rules were found for " + parent.getOperator());

}
/**
 * Evaluate the rules over 2 XATOperators.  
 * <pr>
 * This method uses java.lang.reflect methods to instantiate the classes
 * in this.classVector.  The method iterates over the classes, instantiates the class,
 * then calls the visit function with the operators of parent and child.
 *
 * If no rule is found, the next class is instantiated and the process repeats itself.  
 * the function returns true if the function cannot be pushedDown further
 * otherwise it returns false
 * </pr>
 * @return ReWriteResult An object containing a summary of what was performed by the rule.
 * @param parent The parent node that we will try to find a rule for.
 * @param child The child node that we will try to find a rule for.
 * @throws ReWriteRulesNoRulesFound Exception thrown when the system has looked at
 * all classes in the classVector and is unable to find a rule for parent and child.
 * @since 1.0
 * @see ReWriteResult
 */

public ReWriteResult evaluateRules(XATNode parent, XATNode child)
    throws ReWriteRulesNoRulesFound {

    String className;
    Class ruleClass;
    ReWriteRules p1;
    /*
            System.out.println(
                "Trying to find a rule for: "
                    + parent.getOperator().toString()
                    + " and "
                    + child.getOperator().toString());
    */
    className = parent.getOperator().getClass().toString();

    for (int x = 0; x < classVector.size(); x++) {

        try {
            className = (String) classVector.elementAt(x);
            ruleClass = Class.forName(className);
            //create a new instance of the class from the classVector
            p1 = (ReWriteRules) ruleClass.newInstance();
            //set the variableTable of the new rule Class
            p1.setVariableTable(this.variableTable);
            //set the debug value for each new rule Class
            p1.setDebug(this.isDebug());
            //call the visit(Object, Object) of that class
            result = (ReWriteResult) p1.visit(parent.getOperator(), child.getOperator());
            //return the result if it is not null
            if (result != null) {
                //make the last rule class looked at the first class to make it faster
                String lastClass = (String) classVector.elementAt(x);
                classVector.removeElement(lastClass);
                classVector.insertElementAt(lastClass, 0);
                return result;
            } else {
                //otherwise look at the next class
            }
        } catch (Exception e) {
            System.out.println(e.toString());
        }
    }

    throw new ReWriteRulesNoRulesFound(
        "No rules were found for "
            + parent.getOperator()
            + " and "
            + child.getOperator());

}
/**
 * Find a Navigate operator whose binding matches the entryPoint array.
 * This method accepts a NavigationStep[] entryPoint 
 * and will find a Navigate operator whose binding matches the entryPoint array.
 * This is used to match a Navigate with a Navigate and is useful
 * when trying to determine which side of a join a Navigate matches
 * A Navigate's entryPoint is to be matched with the binding of another  
 * @return boolean True: if a match is found. False: otherwise.
 * @param entryPoint The EntryPoint to match.
 * @param root The root node to start looking at.
 * @since 1.0
 * @see VarStep
 */
protected boolean findMatchingNavigationNode(
    VarStep entryPoint,
    XATNode root) {

    XATNode currentNode;

    String toMatchPath = entryPoint.getStepValue();
    //check the root first
    currentNode = root;
    //check for a source
    if (currentNode.getOperator() instanceof Source) {
        Source currentSource = (Source) currentNode.getOperator();
        if (toMatchPath.equals(currentSource.getOutputColumnName())) {
            return true;
        }
    }
    //check for a Navigate
    if (currentNode.getOperator() instanceof Navigate) {

        Navigate currentNav = (Navigate) currentNode.getOperator();
        //create a string corresponding to the destination
        String childBinding = currentNav.getBinding();
        //if the entryPoint matches the destination
        if (toMatchPath.equals(childBinding)) {
            return true;
        }
    }

    Vector childrenVector = new Vector();
    getChildren(childrenVector, root);

    //loop recursively until there are no children of a node
    if (childrenVector.size() == 0) {
        //if there are no children, do nothing
        return false;

    } else {
        //call this method on each of the children
        for (int x = 0; x < childrenVector.size(); x++) {
            currentNode = (XATNode) childrenVector.elementAt(x);
            if (findMatchingNavigationNode(entryPoint, currentNode)) {
                return true;
            } else {
                //just keep going
            }
        }

    }
    return false;
}
/**
 * An alternate definition for findMatchingNavigationNode.
 * This method takes the binding parameter, converts it to a NavigationStep[] and calls
 * the other findMatchingNavigationNode method.
 * @return boolean True if a match is found. False: otherwise.
 * @param binding The binding of a Navigate operator to the match.
 * @param root The root node to start looking at.
 * @since 1.0
 * @see VarStepImp
 * @see findMatchingNavigationNode(VarStep, XATNode) findMatchingNavigationNode
 */
protected boolean findMatchingNavigationNode(String binding, XATNode root) {

    VarStep ep = new VarStepImp(binding);
    return findMatchingNavigationNode(ep, root);

}
/**
 * Get all children of a given root node.
 * This method accepts a destination Vector and an XATNode and 
 * all of the root's children are puts into destinationVector
 * This is used to convert the root's children (stored as arrays) into a Vector.
 * @return void
 * @param destinationVector The vector to hold all children nodes.
 * @param root The root XATNode to get the children from.
 * @since 1.0
 * @see XATNode
 */
protected void getChildren(Vector destinationVector, XATNode root) {

    XATNode[] children = root.getChildren();
    int x = 0;
    //empty the destinationVector to ensure that only the root's children are in the Vector
    destinationVector.removeAllElements();
    for (x = 0; x < children.length; x++) {

        destinationVector.addElement((XATNode) children[x]);
    }
}
/**
 * Iterates over the BinExpression in order to find all of the column Names
 * anytime a column name is found, it is added into the Vector.
 * 
 * @return boolean
 * @param exp Expression
 */
protected void getColumnNames(BinExpression exp, Vector columnNames) { 
    //check for a terminal expression
    if (exp instanceof TerminalExpression) {
        TerminalExpression te1 = (TerminalExpression) exp;
        //check for a column name terminal expression
        if (te1.getType() == te1.NAME) {
            columnNames.addElement((String)((ColumnName)te1.getValue()).getColumnName());            
        }

    } else {
        try {
            //try to call this on the left
            getColumnNames((BinExpression) exp.getLeft(), columnNames);            
        } catch (ClassCastException e) {
            //if there is a CCE, then the left is a TerminalExpression
            TerminalExpression te1 = (TerminalExpression) exp.getLeft();
            //check for a column name terminal expression
            if (te1.getType() == te1.NAME) {
                columnNames.addElement((String)((ColumnName)te1.getValue()).getColumnName());
            	
            }
        }
            try {
                getColumnNames((BinExpression) exp.getRight(), columnNames);                
            } catch (ClassCastException e1) {
                //same thing as above
                TerminalExpression te1 = (TerminalExpression) exp.getRight();
                //check for a column name terminal expression
                if (te1.getType() == te1.NAME) {
 		            columnNames.addElement((String)((ColumnName)te1.getValue()).getColumnName());	            	
                }

            }
        
    }
    return ;
}
/**
 * Return the statistics of the rewrite rules.
 * @since 1.0  
 */
  
public static java.util.Hashtable getStatis() {
	return statis;
}
/**
 * Insert the method's description here.
 * 
 * @return BindingTable
 */
public BindingTable getVariableTable() {
	return variableTable;
}
/**
 * Get the value of the propertity 'DEBUG_NavTaggerCancelout'.
 * @return boolean True: With debug information. False: No debug information.
 * @since 1.0
 */
protected boolean isDebug() {
	return System.getProperty("DEBUG_ReWriteRules")!=null;
}
/**
 * A flag to check whether to get the statistics from rewrite rules.
 * Creation date: (9/16/2002 5:13:23 PM)
 * @return boolean
 */
protected boolean isGatherStatistics() {
	return System.getProperty("ReWriteRules_GatherStatistics")!=null;
}
/**
 * traverse up the tree, starting from the startingPoint parameter, 
 * checking if columnToFind is used
 * in any entry point in a Navigate, parameter in a Select, or
 * column in a Tagger.
 * It will be failed if the variable is not used in the inner tree, but
 * used in the outer tree. In that case, for the FOR node in the tree,
 * the Sub tree's root Node should have a link to the FOR/GB Node.
 * This bug should be fixed in ParserTraverse. This bug is identified as
 * BUG 698225 on Source Forge.net.
 * Creation date: (7/25/2002 12:50:10 AM)
 * @param column java.lang.String
 */
protected boolean isUsedAbove(XATNode startingPoint, String columnToFind) {

    XATNode n1 = startingPoint;
    XATNode parentNode;
    XATOperator x1;

    if(columnToFind == null)
    	return false;

    while (true) {
        parentNode = n1.getParent();
        //if the parent is not null
        if (parentNode != null) {
            x1 = parentNode.getOperator();
            //check if x1 has a Navigate attached to it
            if (x1 instanceof Navigate) {
                NavigationStep[] entryPoint = ((Navigate) x1).getEntryPoint();
                for (int i = 0; i < entryPoint.length; i++) {
                    String stepValue = entryPoint[i].getStepValue();
                    if (stepValue.equals(columnToFind)) {
                        return true;
                    }
                }
            } else if (x1 instanceof Tagger) { //if not, then check if x1 has a Tagger

                return ((Tagger) x1).getPattern().containsColumn(columnToFind);
                
            } else if (x1 instanceof SelectImp) { //try the Select
                Vector columnNames = new Vector();
                getColumnNames((BinExpression) ((SelectImp) x1).getExpression(), columnNames);
                if (columnNames.contains(columnToFind)) {
                    return true;
                }
            } else if (x1 instanceof Expose) {
	            if (columnToFind.equals(((Expose)x1).getColumnName()))
		            return true;
            } else if (x1 instanceof Aggregate) {
	            if (columnToFind.equals(x1.getOutputColumnName()))
		            return true;
            } else if (x1 instanceof ColumnNameList) { //try the MultiSource operators, e.g., XMLUnion, GroupBy.
                String[] columnNames = ((ColumnNameList) x1).getColumnNames();
                for (int i = 0; i < columnNames.length; i++) {
                    if (columnNames[i].equals(columnToFind)) {
                        return true;
                    }
                }
            } else if(x1 instanceof UpdateOperator){
	            String context = ((UpdateOperator)x1).getContext();
	            String variable = ((UpdateOperator)x1).getVariable();

	            if (columnToFind.equals(context) || columnToFind.equals(variable))
	            	return true;

	            if(x1 instanceof Replace){
		            String replace = ((Replace)x1).getReplacingVariable();
		            if(columnToFind.equals(replace))
		            	return true;
	            }
            } else if( x1 instanceof Function){
				Object[] params = ((Function)x1).getParams();
				//check if the columnToFind is used in param
				for (int i = 0; i < params.length; i++) {
					if (params[i] instanceof ColumnName){
						if (((ColumnName)params[i]).getColumnName().equals(columnToFind)){
							return true;
						}							
					}
				}
            }
            //make n1 = its parent to recursively climb the tree
            n1 = parentNode;
        }
        //the parent was null so we reached the top of the tree, we can just return false
        else {
            return false;
        }
    }
}
    /**
     * <pre>
     * MoveNode will alter the XATTree structue by moving the parent
     * node directly under the child. The child's children will be altered to
     * point to parent. This is primarily used to move a node under the correct Join / Union branch.
     *
     * <b> Structure of tree </b>
     * parent = x2, child = x4
     *   
     *   	old structure
     *   	x1 -> x2 -> x3 -> x4
     *   
     *   	new structure
     *   	x1 -> x3 ->	x2 -> x4
     * 
     * Any other time you want to switch the positions of 2 nodes, swap(XATOperator, XATOperator) 
     * should be used.
     * <pre>
     * @param parent   
     * @param child 
     * @see #swap(XATOperator, XATOperator) swap
     */
    protected void moveNode(XATNode parent, XATNode child) {

        XATNode x1 = parent.getParent();
        XATNode x2 = parent;
        XATNode x3 = child.getParent();
        XATNode x4 = child;

        //disconnect children
        x1.deleteChild(x2);
        x2.deleteChild(x3);
        x3.deleteChild(x4);

        //add the new children
        x1.addChild(x3);
        x3.addChild(x2);
        x2.addChild(x4);

        //add the new parents		
        x3.setParent(x1);
        x2.setParent(x3);
        x4.setParent(x2);

    }
  /**
     * <pre>
     * MoveNodeToRoot will alter the XATTree structue by moving the 
     * node to the root.  This is to be used in conjunction with
     * rename push down.  Renames will be put as the new root,
     * then pushed down and each node it goes through will be
     * changed.
     *
     * <b> Structure of tree </b>
     * root = x1, node = x4
     *   
     *   	old structure
     *   	x1 -> ... -> x3 -> x4 -> x5
     *   
     *   	new structure
     *   	x4 -> x1 -> ... x3 -> x5
     * 
     * <pre>
     * @param root
     * @param nodeToMove
     * @see #visit(Rename, Navigate) rename
     */
protected void moveNodeToRoot(XATNode root, XATNode nodeToMove) {
	
        XATNode x1 = root;
        XATNode x3 = nodeToMove.getParent();
        XATNode x4 = nodeToMove;
        XATNode x5 = null;
        try{
            x5 = x4.getChild(0);
        }catch (XATNodeDoesNotExistException e){
	        System.out.println(e);
        }
        
        //disconnect children
        x3.deleteChild(x4);
	    x4.deleteChild(x5);
	    
        //add the new children
        x4.addChild(x1);
        x3.addChild(x5);

        //add the new parents		
        x1.setParent(x4);
        x5.setParent(x3);
        x4.setParent(null);

}
/**
 * Insert the method's description here.
 * Creation date: (9/16/2002 5:13:23 PM)
 * @param newDebug boolean
 */
public void setDebug(boolean newDebug) {
	if(newDebug)
		System.setProperty("DEBUG_ReWriteRules", "OK");
}
/**
 * Insert the method's description here.
 * 
 * @param newVariableTable java.util.BindingTable
 */
public void setVariableTable(BindingTable newVariableTable) {
	variableTable = newVariableTable;
}
    /**
     *  
     * Swaps the operators of 2 nodes. This is used when the appropriate equivalance rule
     * determines that the 2 nodes can be swapped. 
     * @param operator1 
     * @param operator2 
     */
    protected void swap(XATOperator operator1, XATOperator operator2) {

        XATNode firstNode = operator1.getNode();
        XATNode secondNode = operator2.getNode();

        firstNode.setOperator(operator2);
        secondNode.setOperator(operator1);

        operator1.setNode(secondNode);
        operator2.setNode(firstNode);

    }
/** If there is an inheretance between the visited objects, it will only match the highest class. */
public Object visit(Object object) {
    if (object == null)
        return dispatch();
    Method m = getMethod(object.getClass());
    try {
        if (isGatherStatistics()) {
            if (this.statis == null) {
                this.statis = new Hashtable();
            } else {
                Integer count = (Integer) this.statis.get(m);
                if (count == null) {
                    this.statis.put(m, new Integer(1));
                } else {
                    this.statis.put(m, new Integer(count.intValue() + 1));
                }
            }
        }

        return m.invoke(this, new Object[] { object });
    } catch (InvocationTargetException e) {
        e.getTargetException().printStackTrace();
        return dispatch(object);
    } catch (IllegalAccessException e) {
        return dispatch(object);
    } catch (IllegalArgumentException e) {
        return dispatch(object);
    }
}
    /**
    * We add funtionality to print out the rules used in this rewrite.
    * If there is an inheretance between the visited objects, 
    * it will only match the highest class. 
    */
    public Object visit(Object obj1, Object obj2)
    {
        if(obj1 == null && obj2 == null)
          return dispatch();

        if(obj1 == null)
          return visit(obj2);

        if(obj2 == null)
          return visit(obj1);

        Method m = getMethod(obj1.getClass(), obj2.getClass());
        if(m == null){
            // then get the dispatch object.
            try
            {
                m = getClass().getMethod("dispatch",
                    new Class[] { Object.class, Object.class });
            }
            catch (NoSuchMethodException nsme)
            {
                System.out.println("Cannot Happend!");
            }
        }
        try{
	        if(isGatherStatistics()){
		        if(this.statis == null){
			        this.statis = new Hashtable();
		        }else{
			        Integer count = (Integer)this.statis.get(m);
			        if(count == null){
				        this.statis.put(m, new Integer(1));
			        }else{
				        this.statis.put(m, new Integer(count.intValue()+1));
			        }
		        }
	        }
            return m.invoke(this, new Object[] { obj1, obj2 });
        }catch(InvocationTargetException e){
          return dispatch(obj1, obj2);
        }catch(IllegalAccessException e){
          return dispatch(obj1, obj2);
        }catch(IllegalArgumentException e){
          return dispatch(obj1, obj2);
        }
     }
}
